/*****************************************************************************

	GTK Development Environment
	
	Copyright (c) 2000 End of the World Software

	date:	December 05 2000
	author: Gertjan Zwartjes
	email:	g.zwartjes@student.tue.nl

  **************************************************************************

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gde.h"


/*=========================================================

	Variable Declarations

=========================================================*/
GdeProject *project = NULL;


/*---------------------------------------------------------

	project_find_node()

	Find a node in the project, given a specific node.
	Returns a pointer to th file list entry.

---------------------------------------------------------*/
GdeFileList *project_find_node(GtkCTreeNode *node)
{
	GdeFileList *file_list;
	
	file_list = file_list_find_node(project->source_list, node);
	if (file_list == NULL)
		file_list = file_list_find_node(project->header_list, node);
	if (file_list == NULL)
		file_list = file_list_find_node(project->root_list, node);

	return file_list;
}


/*----------------------------------------------------------

	project_modified()

	Checks all file lists, if they are modified or not.

----------------------------------------------------------*/
gboolean project_modified(void)
{
	return file_list_modified(project->source_list)
		|| file_list_modified(project->header_list) 
		|| file_list_modified(project->root_list);
}


/*----------------------------------------------------------

	project_modified_reset()

	Reset all file_list modified attribute.

----------------------------------------------------------*/
void project_modified_reset(void)
{
	file_list_modified_reset(project->source_list);	
	file_list_modified_reset(project->header_list);	
	file_list_modified_reset(project->root_list);	
}


/*---------------------------------------------------------

	project_delete()

	Frees up memory took by project.

---------------------------------------------------------*/
void project_delete(void)
{
	file_list_clear(project->source_list);
	file_list_clear(project->header_list);
	file_list_clear(project->root_list);
	
	g_free(project->source_list);
	g_free(project->header_list);
	g_free(project->root_list);

	g_free(project);
}


/*---------------------------------------------------------

	project_add_to_project()

	Add a file to the current project. Type identifies
	the type of file. If 't' test on file extension;
	's' source, 'h' header, 'r' root.
	
---------------------------------------------------------*/
void project_add_to_project(const char *filename, char type)
{
	GdeFileList *file_list = NULL;	
	gchar *short_name = g_basename(filename);

	if (type == 't')
		type = type_from_filename(filename);
	
	switch(type) {
	case 's':
		file_list = file_list_new_item(project->source_list);
		file_list->tree_item = proj_win_add_source_file(short_name);
		break;
	case 'h':
		file_list = file_list_new_item(project->header_list);
		file_list->tree_item = proj_win_add_header_file(short_name);
		break;
	case 'r':
		file_list = file_list_new_item(project->root_list);
		file_list->tree_item = proj_win_add_root_file(short_name);
	}
	
	file_list->filename = g_strdup(filename);

	project->needs_save = TRUE;
	main_win_set_title(g_basename(project->filename));
}


/*---------------------------------------------------------

	project_add()

	Opens the file selector to add a file to the
	project.

---------------------------------------------------------*/
void project_add(void)
{
	GList *file_list;

	file_list = file_win("Add File(s) to Project", FILE_WIN_SELECT_MULTIPLE);

	if (file_list) {
		while (file_list) {
			project_add_to_project(file_list->data, 't');
			g_free(file_list->data);
			file_list = g_list_next(file_list);
		}

		g_list_free(file_list);
	}
}


/*---------------------------------------------------------

	project_remove()

	Removes the currently selected file.

---------------------------------------------------------*/
void project_remove(void)
{
	GtkCTreeNode *node;
	GdeFileList *file_list;
	
	node = proj_win_selected_node();

	if (node) {
		file_list = project_find_node(node);
		if (file_list) {
			if (file_list->prev_file)
				((GdeFileList*) file_list->prev_file)->next_file =
													file_list->next_file;
			if (file_list->next_file)
				((GdeFileList*) file_list->next_file)->prev_file = 
													file_list->prev_file;
			g_free(file_list);
			proj_win_remove_file(node);

			project->needs_save = TRUE;
			main_win_set_title(g_basename(project->filename));
		}
	}

}


/*---------------------------------------------------------

	project_execute_command()

	Execute the given command. The command's output pipe
	is redirected to the status window.
	
---------------------------------------------------------*/
void project_execute_command(const gchar* command)
{
	int out_pipe[2], err_pipe[2], flags;
	gboolean read;
	FILE *out, *err;
	pid_t pid;
	int c;
	gchar str[512];

	if ((command) && (*command) && !project->command_running_pid) {
		sprintf(str, "Executing: %s\n", command);
		status_win_text_insert(str, STATUS_TEXT_INFO);
		
		if (pipe(out_pipe))
			g_error("Could not create pipe: %s", g_strerror(errno));
		if (pipe(err_pipe))
			g_error("Could not create pipe: %s", g_strerror(errno));
			
		pid = fork();
		if (pid == -1) {
			g_error("Could not create process: %s", g_strerror(errno));
		}
		else if (pid == 0) {
			close(STDOUT_FILENO); dup(out_pipe[1]);
			close(STDERR_FILENO); dup(err_pipe[1]);
			if (execl("/bin/sh", "/bin/sh", "-c", command, NULL) == -1)
				g_error("Could not execute shell: %s", g_strerror(errno));
			_exit(0);
		}
		else {
			project->command_running_pid = pid;
			main_win_disable_commands();
			
			close(out_pipe[1]);
			close(err_pipe[1]);
			out = fdopen(out_pipe[0], "r");
			err = fdopen(err_pipe[0], "r");

			/* set nonblocking file mode else we will lockup */
			flags = fcntl(out_pipe[0], F_GETFL, 0);
			flags |= O_NONBLOCK;
			fcntl(out_pipe[0], F_SETFL, flags);
			flags = fcntl(err_pipe[0], F_GETFL, 0);
			flags |= O_NONBLOCK;
			fcntl(err_pipe[0], F_SETFL, flags);
			read = FALSE;
			
			if (out) {
				while (!feof(out) || !feof(err)) {
					if ((c = fgetc(out)) != EOF) {
						if (c != 255) {
							status_win_char_insert(c, STATUS_TEXT_OUTPUT);
							read = TRUE;
						}
					}
					if ((c = fgetc(err)) != EOF) {
						if (c != 255) {
							status_win_char_insert(c, STATUS_TEXT_ERROR);
							read = TRUE;
						}
					}
					if (!read) {
						if (gtk_events_pending()) {
							gtk_main_iteration_do(FALSE);
						}						
						else {
							usleep(50);
						}
					}
					read = FALSE;
				}
				fclose(out);
				fclose(err);
			}

			close(out_pipe[0]);
			close(err_pipe[0]);

			main_win_enable_commands();
			project->command_running_pid = 0;
			
			status_win_text_insert("Ready\n\n", STATUS_TEXT_INFO);
		}
	}
}


/*----------------------------------------------------------

	project_kill_command()

	Kills the current running command.

----------------------------------------------------------*/
void project_kill_command(void)
{
	if (project->command_running_pid) {
		if(kill(project->command_running_pid, SIGKILL) == -1)
			g_warning("Could not kill process %i: %s", project->command_running_pid, g_strerror(errno));
	}
}


/*---------------------------------------------------------

	project_make()

	Execute the make command for the project.
	
---------------------------------------------------------*/
void project_make(void)
{
	if (project->check_dates)
		project_modified_reset();
	project_execute_command(project->make_command);
}


/*---------------------------------------------------------

	project_run()

	Execute the run command for the current project.

---------------------------------------------------------*/
void project_run(void)
{
	MessageWinButton mwb = MESSAGE_BUTTON_NO;

	if (project->check_dates && project_modified()) {
		mwb = message_win("Run", "Some files have been changed since last make!\nDo you want to execute make command first?");
		if (mwb == MESSAGE_BUTTON_YES) {
			project_make();
		}
	}
	
	if (mwb != MESSAGE_BUTTON_CANCEL)
		project_execute_command(project->run_command);
}


/*---------------------------------------------------------

	project_clean()

	Executes the clean command for the active project.

---------------------------------------------------------*/
void project_clean(void)
{
	project_execute_command(project->clean_command);
}


/*---------------------------------------------------------

	project_edit_file()

	Opens the file corresponding to the node clicked.

---------------------------------------------------------*/
void project_edit(void)
{
	GtkCTreeNode *node;
	GdeFileList *file_list;
	pid_t pid;
	gchar command[512];
	
	if (project->edit_command) {
		node = proj_win_selected_node();
		
		if (node) {
			file_list = project_find_node(node);

			if (file_list) {
				sprintf(command, "%s %s", project->edit_command, file_list->filename);
				if (file_list) {
					pid = fork();
					if (pid == -1) {
						g_error("Could not create process: %s", g_strerror(errno));
					}
					else if (pid == 0) {
						if (execl("/bin/sh", "/bin/sh", "-c", command, NULL) == -1)
							g_error("Could not execute shell: %s", g_strerror(errno));
						_exit(0);
					}
				}
			}
		}
	}
}


/*---------------------------------------------------------

	project_new()

	Creates a new empty project.

---------------------------------------------------------*/
void project_new(void)
{
	if (!project) {
		project = g_malloc0(sizeof(GdeProject));
		project->source_list = file_list_new();
		project->header_list = file_list_new();
		project->root_list = file_list_new();
	}

	/* clear all values of project struct */
	project_close();
}


/*---------------------------------------------------------

	project_save()

	Saves the project data to file.

---------------------------------------------------------*/
void project_save(void)
{
	FILE *file;

	if (*project->filename) {
		file = fopen(project->filename, "w");

		if (file) {
			fprintf(file, "%c Gtk Development Environment %s Project File\n\n", 
					COMMENT, GDE_VERSION);
			fprintf(file, "%c General project settings\n", COMMENT);
			fprintf(file, "edit_command=%s\n", project->edit_command);
			fprintf(file, "make_command=%s\n", project->make_command);
			fprintf(file, "run_command=%s\n", project->run_command);
			fprintf(file, "clean_command=%s\n", project->clean_command);
			
			fprintf(file, "check_dates=%i\n", project->check_dates);
	
			fprintf(file, "\n%c Source Files\n", COMMENT);
			file_list_save(file, project->source_list, "srcfile=");
			fprintf(file, "\n%c Header Files\n", COMMENT);
			file_list_save(file, project->header_list, "hdrfile=");
			fprintf(file, "\n%c Root Files\n", COMMENT);
			file_list_save(file, project->root_list, "rootfile=");

			fclose(file);

			project->needs_save = FALSE;
			main_win_set_title(g_basename(project->filename));
		}
		else
			g_warning("Could not open '%s' for writing", project->filename);
	}
	else {
		project_save_as();
	}
}


/*---------------------------------------------------------

	project_save_as()

	Opens a file selecter and saves the project as the
	chosen file name.

---------------------------------------------------------*/
void project_save_as(void)
{
	GList *file_list;
	gchar *dirname;

	file_list = file_win("Save Project As", FILE_WIN_SELECT_SINGLE);

	if (file_list) {
		strcpy(project->filename, file_list->data);
		
		main_win_set_title(g_basename(file_list->data));
		dirname = g_dirname(file_list->data);
		chdir(dirname);
		g_free(dirname);
		
		main_win_recent_add(file_list->data);
			
		project_save();

		g_free(file_list->data);
		g_list_free(file_list);
	}
}


/*---------------------------------------------------------

	project_close()

	Empties the current project. To `no' values.

---------------------------------------------------------*/
gboolean project_close(void)
{
	MessageWinButton mwb;

	if (project->needs_save)
		mwb = message_win(g_basename(project->filename), "Do you want to save changes?");
	else
		mwb = MESSAGE_BUTTON_NO;
	
	if (mwb != MESSAGE_BUTTON_CANCEL) {
		if (mwb == MESSAGE_BUTTON_YES) {
			project_save();
		}
		/* Last opened is also last closed! */
		strcpy(gde_rc->last_opened, project->filename);
		project->filename[0] = 0;
	
		project->edit_command[0] = 0;
		project->make_command[0] = 0;
		project->clean_command[0] = 0;
		project->run_command[0] = 0;
		project->command_running_pid = 0;

		project->check_dates = TRUE;
		project->needs_save = FALSE;
	
		file_list_clear(project->source_list);
		file_list_clear(project->header_list);
		file_list_clear(project->root_list);

		main_win_disable_commands();
		main_win_enable_commands();
		main_win_set_title(NULL);
		status_win_clear();

		return TRUE;
	}
	else
		return FALSE;
}
	

/*---------------------------------------------------------

	project_open_file()

	Opens the specific file into GDE. If the file is not
	found, a new project is created with filename as
	filename.

---------------------------------------------------------*/
void project_open_file(const gchar *filename)
{
	gchar buffer[512];
	gchar *dirname;
	FILE *file;

	gint read = 0;
	
	if (filename) {
		strcpy(project->filename, filename);
		file = fopen(filename, "r");
		if (file) {
			sprintf(buffer, "Opening %s\n", filename);
			status_win_text_insert(buffer, STATUS_TEXT_INFO);
			
			while (!feof(file)) {
				memset(buffer, 0, sizeof(buffer));
				fgets(buffer, sizeof(buffer)-1, file);
				g_strchug(buffer);

				if ((*buffer) && (buffer[0] != COMMENT)) {
					if (str_get_value_of("edit_command=", buffer)) {
						read |= 0x01;
						strcpy(project->edit_command, buffer);
					}
					if (str_get_value_of("make_command=", buffer)) { 
						read |= 0x02;
						strcpy(project->make_command, buffer);
					}
					if (str_get_value_of("clean_command=", buffer)) {
						read |= 0x04;
						strcpy(project->clean_command, buffer);
					}
					if (str_get_value_of("run_command=", buffer)) {
						read |= 0x08;
						strcpy(project->run_command, buffer);
					}
					if (str_get_value_of("check_dates=", buffer)) {
						read |= 0x10;
						project->check_dates = strtol(buffer, NULL, 10);
					}
			
					if (str_get_value_of("srcfile=", buffer)) {
						read |= 0x20;
						project_add_to_project(buffer, 's');
					}
					if (str_get_value_of("hdrfile=", buffer)) {
						read |= 0x40;
						project_add_to_project(buffer, 'h');
					}
					if (str_get_value_of("rootfile=", buffer)) {
						read |= 0x80;
						project_add_to_project(buffer, 'r');
					}
				}
			}
			fclose(file);

			project->needs_save = FALSE;
			main_win_set_title(g_basename(filename));

			dirname = g_dirname(filename);
			chdir(dirname);
			g_free(dirname);
				
			if (project->check_dates)
				project_modified_reset();

			main_win_enable_commands();
			main_win_recent_add(filename);

			if (read & 0x1F)
				status_win_text_insert("Loaded commands\n", STATUS_TEXT_OUTPUT);
				
			if (!(read & 0x01))
				status_win_text_insert("No edit command found\n", STATUS_TEXT_ERROR);
			if (!(read & 0x02))
				status_win_text_insert("No make command found\n", STATUS_TEXT_ERROR);
			if (!(read & 0x04))
				status_win_text_insert("No clean command found\n", STATUS_TEXT_ERROR);
			if (!(read & 0x08))
				status_win_text_insert("No run command found\n", STATUS_TEXT_ERROR);
			if (!(read & 0x10))
				status_win_text_insert("No check dates option found\n", STATUS_TEXT_ERROR);

			if (read & 0x20)
				status_win_text_insert("Loaded source files\n", STATUS_TEXT_OUTPUT);
			if (read & 0x40)
				status_win_text_insert("Loaded header files\n", STATUS_TEXT_OUTPUT);
			if (read & 0x80)
				status_win_text_insert("Loaded root files\n", STATUS_TEXT_OUTPUT);
			
			status_win_text_insert("Ready\n\n", STATUS_TEXT_INFO);
		}
		else if (*filename) {
			sprintf(buffer, "New project: %s\n", filename);
			status_win_text_insert(buffer, STATUS_TEXT_INFO);
			status_win_text_insert("Ready\n\n", STATUS_TEXT_INFO);
		}
	}
}


/*---------------------------------------------------------

	project_open()

	Opens the file selector and loads the specific 
	project out of the file.

---------------------------------------------------------*/
void project_open(void)
{
	GList *file_list;
	
	if (project_close()) {
		file_list = file_win("Open Project", FILE_WIN_SELECT_SINGLE);

		if (file_list) {
			project_open_file(file_list->data);

			g_free(file_list->data);
			g_list_free(file_list);
		}
	}
}
